// Copyright 2012 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// reduq December 2012 version

#include "reducible_queue.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

namespace {

class OpaqueInt {
 public:
  explicit OpaqueInt(int x) : x_(x) {}
  int x() const { return x_; }
 private:
  int x_;
};

bool operator==(const OpaqueInt& x, const OpaqueInt& y) {
  return x.x() == y.x();
}

struct Left {
  typedef OpaqueInt Type;
  static OpaqueInt reduce(OpaqueInt x, OpaqueInt /*y*/) { return x; }
  static OpaqueInt id() { return OpaqueInt(0); }
};

void FirstForest(int f[], int n) {
  for (int i(n); i != -1; --i) f[i] = 2 * i + 1;
}

bool NextForest(int f[], int n) {
  int i(0);
  while (f[i] + 1 == f[i + 1]) ++i;
  ++f[i];
  FirstForest(f, i - 1);
  return i + 1 < n;
}

template<typename Reducer>
void TestQueue() {
  typedef typename Reducer::Type T;
  enum { N = 11 };
  int f[N + 1];
  FirstForest(f, N);
  do {
    int i(0);
    reduq::reducible_queue<Reducer> q;
    std::list<T> l;
    for (int j(0); j != 2 * N; ++j) {
      if (j != f[i]) {
        if (false) putc('(', stderr);
        T x(rand() - rand());
        assert(x == Reducer::reduce(x, Reducer::id()));
        q.push(x);
        l.push_back(x);
      } else {
        if (false) putc(')', stderr);
        ++i;
        q.pop();
        l.pop_front();
      }
      assert(q.empty() == l.empty());
      assert(q.size() == l.size());
      if (!l.empty()) {
        assert(q.front() == l.front());
        assert(q.back() == l.back());
      }
      T ans(Reducer::id());
      for (typename std::list<T>::reverse_iterator k(l.rbegin());
           k != l.rend();
           ++k) {
        ans = Reducer::reduce(*k, ans);
      }
      assert(q.reduce() == ans);
    }
    if (false) putc('\n', stderr);
  } while (NextForest(f, N));
}
}  // namespace

int main() {
  TestQueue<Left>();
  TestQueue<reduq::BitwiseAnd<unsigned> >();
  TestQueue<reduq::BitwiseOr<unsigned> >();
  TestQueue<reduq::BitwiseXor<unsigned> >();
  TestQueue<reduq::IntegerMax<int> >();
  TestQueue<reduq::IntegerMin<int> >();
  TestQueue<reduq::Product<unsigned> >();
  TestQueue<reduq::RealMax<double> >();
  TestQueue<reduq::RealMin<double> >();
  TestQueue<reduq::Sum<unsigned> >();
  fputs("OK\n", stderr);
}
